"""Pylint plugin: test classes derived from test classes."""

import astroid

from pylint.checkers import BaseChecker, utils
from pylint.interfaces import IAstroidChecker

from .common import BASE_ID, check_visitors


def register_checkers(linter):
    """Register checkers."""
    linter.register_checker(LayeredTestClassChecker(linter))


def is_test_case_class(node):
    """Is this node a test class?

    To be a test class, it has to derive from unittest.TestCase, and not
    have __test__ defined as False.

    """
    if not node.is_subtype_of("unittest.case.TestCase"):
        return False

    dunder_test = node.locals.get("__test__")
    if dunder_test:
        if isinstance(dunder_test[0], astroid.AssignName):
            value = list(dunder_test[0].assigned_stmts())
            if len(value) == 1 and isinstance(value[0], astroid.Const):
                return bool(value[0].value)

    return True


@check_visitors
class LayeredTestClassChecker(BaseChecker):
    """Pylint checker for tests inheriting test methods from other tests."""

    __implements__ = (IAstroidChecker,)

    name = "layered-test-class-checker"

    MESSAGE_ID = "test-inherits-tests"
    msgs = {
        ("E%d03" % BASE_ID): (
            "test class %s inherits tests from %s",
            MESSAGE_ID,
            "Used when a test class inherits test methods from another test "
            "class, meaning the inherited tests will run more than once.",
        )
    }

    @utils.check_messages(MESSAGE_ID)
    def visit_classdef(self, node):
        """Check each class."""
        if not is_test_case_class(node):
            return

        for anc in node.ancestors():
            if not is_test_case_class(anc):
                continue
            for meth in anc.mymethods():
                if meth.name.startswith("test_"):
                    self.add_message(self.MESSAGE_ID, args=(node.name, anc.name), node=node)
                    # No need to belabor the point.
                    return



"""
The logic for writing shared files.
"""

import os
import os.path

import pkg_resources
from six.moves import cStringIO
from six.moves import configparser

from edx_lint import __version__
from edx_lint.configfile import merge_configs
from edx_lint.tamper_evident import TamperEvidentFile
from edx_lint.metadata import KNOWN_FILES


def make_header(metadata, filename, tweaks_name):
    """Construct the appropriate header warning for `filename`.

    (Sorry about the logic here, I didn't want to install Jinja just for this.)
    """

    header = """\
***************************
** DO NOT EDIT THIS FILE **
***************************

This file was generated by edx-lint: https://github.com/openedx/edx-lint

"""
    if tweaks_name:
        header += """\
If you want to change this file, you have two choices, depending on whether
you want to make a local change that applies only to this repo, or whether
you want to make a central change that applies to all repos using edx-lint.

Note: If your {filename} file is simply out-of-date relative to the latest
{filename} in edx-lint, ensure you have the latest edx-lint installed
and then follow the steps for a "LOCAL CHANGE".

LOCAL CHANGE:

1. Edit the local {tweaks_name} file to add changes just to this
   repo's file.

2. Run:

      $ edx_lint write {filename}

3. This will modify the local file.  Submit a pull request to get it
   checked in so that others will benefit.

"""
    else:
        header += """\
If you want to change this file, you must edit the central file in the
edx-lint repo.

Note: If your {filename} file is simply out-of-date relative to the latest
{filename} in edx-lint, ensure you have the latest edx-lint installed
and then run:

      $ edx_lint write {filename}

"""
    header += """\

CENTRAL CHANGE:

1. Edit the {filename} file in the edx-lint repo at
   https://github.com/openedx/edx-lint/blob/master/edx_lint/files/{filename}

2. install the updated version of edx-lint (in edx-lint):

      $ pip install .

3. Run (in edx-lint):

      $ edx_lint write {filename}

4. Make a new version of edx_lint, submit and review a pull request with the
   {filename} update, and after merging, update the edx-lint version and
   publish the new version.

5. In your local repo, install the newer version of edx-lint.

6. Run:

      $ edx_lint write {filename}

7. This will modify the local file.  Submit a pull request to get it
   checked in so that others will benefit.





STAY AWAY FROM THIS FILE!





SERIOUSLY.

------------------------------
Generated by edx-lint version: {version}
------------------------------
"""
    text = header.format(filename=filename, tweaks_name=tweaks_name, version=__version__)
    text = "".join((metadata.comment.format(line).rstrip() + "\n") for line in text.splitlines())
    return text


def write_file(filename, output_fn):
    """Write `filename`.

    Messages go to `output_fn`.  Return value is an exit status for the
    main command to exit with.
    """
    metadata = KNOWN_FILES.get(filename)
    if metadata is None:
        output_fn("Don't have file %r to write." % filename)
        return 2

    if os.path.exists(filename):
        output_fn("Checking existing copy of %s" % filename)
        tef = TamperEvidentFile(filename)
        if not tef.validate():
            bak_name = amend_filename(filename, "_backup")
            output_fn("Your copy of %s seems to have been edited, renaming it to %s" % (filename, bak_name))
            if os.path.exists(bak_name):
                output_fn("A previous %s exists, deleting it" % bak_name)
                os.remove(bak_name)
            os.rename(filename, bak_name)

    output_fn("Reading edx_lint/files/%s" % filename)

    file_content = get_file_content(filename)

    tweaks_name = None
    if metadata.format == "ini":
        cfg = configparser.RawConfigParser()
        cfg.read_string(file_content, f"<internal {filename}>")

        tweaks_name = amend_filename(filename, "_tweaks")
        if os.path.exists(tweaks_name):
            output_fn("Applying local tweaks from %s" % tweaks_name)
            cfg_tweaks = configparser.RawConfigParser()
            cfg_tweaks.read([tweaks_name])

            merge_configs(cfg, cfg_tweaks)

    output_fn("Writing %s" % filename)
    output_text = cStringIO()
    output_text.write(make_header(metadata=metadata, filename=filename, tweaks_name=tweaks_name))
    if metadata.format == "ini":
        cfg.write(output_text)
    else:
        output_text.write(file_content)

    out_tef = TamperEvidentFile(filename)
    output_bytes = output_text.getvalue().encode("utf8")
    out_tef.write(output_bytes, hashline=metadata.comment)

    return 0


def get_file_content(filename):
    """Get the contents of the file that should be written."""
    return pkg_resources.resource_string("edx_lint", f"files/{filename}").decode("utf8")


def amend_filename(filename, amend):
    """Amend a filename with a suffix.

    amend_filename("foo.txt", "_tweak") --> "foo_tweak.txt"

    """
    base, ext = os.path.splitext(filename)
    amended_name = base + amend + ext
    return amended_name



"""pytest setup for this directory."""

import os
import shutil

import pytest


@pytest.fixture(autouse=True)
def run_tests_in_temp_dir(tmpdir):
    """Automatically runs all tests in their own temp directory."""

    cur_dir = os.getcwd()
    new_dir = str(tmpdir)
    os.chdir(new_dir)
    yield
    os.chdir(cur_dir)
    shutil.rmtree(new_dir)



"""The edx_lint command."""

import sys

from edx_lint import __version__
from edx_lint.cmd.check import check_main
from edx_lint.cmd.list import list_main
from edx_lint.cmd.write import write_main
from edx_lint.cmd.update import update_main


def main(argv=None):
    """The edx_lint command entry point."""
    if argv is None:
        argv = sys.argv[1:]

    if not argv or argv[0] == "help":
        show_help()
        return 0
    elif argv[0] == "check":
        return check_main(argv[1:])
    elif argv[0] == "list":
        return list_main(argv[1:])
    elif argv[0] == "write":
        return write_main(argv[1:])
    elif argv[0] == "update":
        return update_main(argv[1:])
    else:
        print("Don't understand {!r}".format(" ".join(argv)))
        show_help()
        return 1


def show_help():
    """Print the help string for the edx_lint command."""
    print(
        """\
Manage local config files from masters in edx_lint.
version {VERSION}
Commands:
""".format(VERSION=__version__)
    )
    for cmd in [write_main, check_main, list_main, update_main]:
        print(cmd.__doc__.lstrip("\n"))



"""A pylint checker that simply records what module it's visiting.

This helps diagnose problems with pylint not running on all files.

To use, define an environment variable PYLINT_RECORD_FILES, with a value of
a file name to write them to:

    set PYLINT_RECORD_FILES=pylinted_files.txt

"""

import os

from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from .common import BASE_ID, check_visitors


FILENAME = os.environ.get("PYLINT_RECORD_FILES", "")


def register_checkers(linter):
    """Register checkers."""
    if FILENAME:
        linter.register_checker(ModuleTracingChecker(linter))


@check_visitors
class ModuleTracingChecker(BaseChecker):
    """
    Not really a checker, it doesn't generate any messages.  There's probably
    a better way to hook into pylint to do this.
    """

    __implements__ = (IAstroidChecker,)

    name = "module-tracing-checker"

    msgs = {("E%d00" % BASE_ID): ("bogus", "bogus", "bogus")}

    def visit_module(self, node):
        """Called for each module being examined."""
        # This pointless statement is here because removing the statement causes a different warning. Currently we run
        # pylint at multiple versions in the Open edX platform.  However, the other warning (no-self-use) exists in the
        # old version of pylint (pre 2.13.0) but not the new version. This is the only way to fix the warning in a way
        # that will work for both the old and new version.
        self    # pylint: disable=pointless-statement
        with open(FILENAME, "a") as f:
            f.write(node.file)
            f.write("\n")



"""Test getattr_check.py"""

from .pylint_test import run_pylint


def test_getattr_checker():
    source = """\
        getattr(name, "hello")                  #=A
        getattr(name, "hello", 17)
        getattr(name, "hello", None)

        setattr(name, "hello", hello)           #=B
        setattr(name, "h" + "ello", world)

        delattr(name, "something")              #=C
        delattr(name, "FOO".lower())

        # You can use a literal if it's not a valid identifier
        getattr(name, "hello-world")
        getattr(name, "hello.world")
        getattr(name, "")
        getattr(name, " ")
        getattr(name, "1x")

        # The warnings can be disabled
        getattr(name, "_")                      # pylint: disable=literal-used-as-attribute
        if 1:
            # pylint: disable=literal-used-as-attribute
            getattr(name, "Hello")
            setattr(name, "hello", "hello")

        # More bad cases
        getattr(name, "hello1")                 #=D
        getattr(name, "_")                      #=E

        # Account for this case in our code...
        world = getattr(name, 1)

        # We don't get confused by another function nname
        foobar(name, "hello")
    """
    msg_ids = "literal-used-as-attribute"
    messages = run_pylint(source, msg_ids)
    expected = {
        "A:literal-used-as-attribute:getattr using a literal attribute name",
        "B:literal-used-as-attribute:setattr using a literal attribute name",
        "C:literal-used-as-attribute:delattr using a literal attribute name",
        "D:literal-used-as-attribute:getattr using a literal attribute name",
        "E:literal-used-as-attribute:getattr using a literal attribute name",
    }
    assert expected == messages



"""
Tests of the lint-amnesty command.
"""

from collections import defaultdict
from io import StringIO
import itertools
import textwrap
import unittest

from edx_lint.cmd.amnesty import PylintError, fix_pylint, PYLINT_EXCEPTION_REGEX, parse_pylint_output


class AmnestyTest(unittest.TestCase):
    """
    Tests of lint-amnesty.
    """

    def assert_amnesty(self, input_code, errors, expected):
        """
        Assert that fix_pylint produces ``expected`` when fed ``input_code`` and the
        list of errors ``errors``.

        Arguments:
            input_code: A string of python code. Will be textwrap.dedented.
            errors: A list of PylintErrors
            expected: A string of python code. Will be textwrap.dedented.
        """
        input_code = textwrap.dedent(input_code)
        expected = textwrap.dedent(expected)

        errors_by_line = defaultdict(list)
        for error in errors:
            errors_by_line[error.linenum].append(error)

        output_lines = itertools.chain.from_iterable(
            fix_pylint(line, errors_by_line[lineno]) for lineno, line in enumerate(StringIO(input_code), start=1)
        )

        self.assertEqual(expected.split("\n"), "".join(output_lines).split("\n"))

    def assert_pylint_exception_match(self, expected, line):
        """
        Assert that PYLINT_EXCEPTION_REGEX mateches ``expected`` in ``line``.
        """
        self.assertEqual(expected, PYLINT_EXCEPTION_REGEX.search(line).group(0))

    def test_pylint_exception_re(self):
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception, other-exception",
            "some line  # pylint: disable=some-exception, other-exception",
        )
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception,other-exception",
            "some line  # pylint: disable=some-exception,other-exception",
        )
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception", "some line  # pylint: disable=some-exception"
        )
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception", "some line  # pylint: disable=some-exception#noqa"
        )
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception", "some line  # pylint: disable=some-exception  #noqa"
        )
        self.assert_pylint_exception_match(
            "# pylint: disable=some-exception", "some line# pylint: disable=some-exception#noqa"
        )

    def test_parse_pylint_output(self):
        pylint_output = textwrap.dedent(
            '''\
            ************* Module cms.urls
            cms/urls.py:1: [C0111(missing-docstring), ] Missing module docstring
            ************* Module cms.envs.test
            cms/envs/test.py:81: [C0301(line-too-long), ] Line too long (144/120)
            ************* Module test.test_amnesty
            test/test_amnesty.py:78: [C0326(bad-whitespace), ] Exactly one space required around assignment
                    pylint_output=textwrap.dedent("""\\
                                 ^
        '''
        )
        self.assertEqual(
            [
                PylintError("cms/urls.py", 1, "C0111", "missing-docstring", "", "Missing module docstring"),
                PylintError("cms/envs/test.py", 81, "C0301", "line-too-long", "", "Line too long (144/120)"),
                PylintError(
                    "test/test_amnesty.py",
                    78,
                    "C0326",
                    "bad-whitespace",
                    "",
                    "Exactly one space required around assignment",
                ),
            ],
            list(parse_pylint_output(pylint_output.split("\n"))),
        )

    def test_add_supression(self):
        input_code = """\
            def func(arg):
                pass
        """

        errors = [PylintError("test.py", 1, "W0613", "unused-argument", "func", "Unused argument 'arg'")]

        expected = """\
            def func(arg):  # lint-amnesty, pylint: disable=unused-argument
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_add_multiple_suppressions_same_line(self):
        input_code = """\
            def func(arg, arg):
                pass
        """

        errors = [
            PylintError("test.py", 1, "W0613", "unused-argument", "func", "Unused argument 'arg'"),
            PylintError("test.py", 1, "E0108", "duplicate-argument-name", "func", "Duplicate argument name 'arg'"),
        ]

        expected = """\
            def func(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name, unused-argument
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_add_to_existing_suppression(self):
        input_code = """\
            def func(arg, arg):  # pylint: disable=unused-argument
                pass
        """

        errors = [
            PylintError("test.py", 1, "E0108", "duplicate-argument-name", "func", "Duplicate argument name 'arg'")
        ]

        expected = """\
            def func(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name, unused-argument
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_add_to_existing_suppression_trailing_comment(self):
        input_code = """\
            def func(arg, arg):  # pylint: disable=unused-argument  # noqa
                pass
        """

        errors = [
            PylintError("test.py", 1, "E0108", "duplicate-argument-name", "func", "Duplicate argument name 'arg'")
        ]

        expected = """\
            def func(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name, unused-argument  # noqa
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_add_suppression_trailing_comment(self):
        input_code = """\
            def func(arg, arg):  # noqa
                pass
        """

        errors = [
            PylintError("test.py", 1, "E0108", "duplicate-argument-name", "func", "Duplicate argument name 'arg'")
        ]

        expected = """\
            def func(arg, arg):  # noqa  # lint-amnesty, pylint: disable=duplicate-argument-name
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_remove_useless_suppression(self):
        input_code = """\
            def func(arg, arg):  # pylint: disable=unused-argument
                pass
        """

        errors = [
            PylintError(
                "test.py", 1, "I0021", "useless-suppression", "func", "Useless suppression of 'unused-argument'"
            )
        ]

        expected = """\
            def func(arg, arg):
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_remove_useless_suppression_leave_rest(self):
        input_code = """\
            def func(arg, arg):  # pylint: disable=unused-argument, duplicate-argument-name
                pass
        """

        errors = [
            PylintError(
                "test.py", 1, "I0021", "useless-suppression", "func", "Useless suppression of 'unused-argument'"
            )
        ]

        expected = """\
            def func(arg, arg):  # pylint: disable=duplicate-argument-name
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_replace_useless_suppression(self):
        input_code = """\
            def func(arg, arg):  # pylint: disable=unused-argument
                pass
        """

        errors = [
            PylintError(
                "test.py", 1, "I0021", "useless-suppression", "func", "Useless suppression of 'unused-argument'"
            ),
            PylintError("test.py", 1, "E0108", "duplicate-argument-name", "func", "Duplicate argument name 'arg'"),
        ]

        expected = """\
            def func(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_fix_multiple_lines(self):
        input_code = """\
            def func_a(arg):
                pass

            def func_b(arg, arg):
                pass
        """

        errors = [
            PylintError("test.py", 1, "W0613", "unused-argument", "func", "Unused argument 'arg'"),
            PylintError("test.py", 4, "E0108", "duplicate-argument-name", "func", "Duplicate argument name 'arg'"),
        ]

        expected = """\
            def func_a(arg):  # lint-amnesty, pylint: disable=unused-argument
                pass

            def func_b(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_fix_module_docstring_missing(self):
        input_code = """\
            def func(arg):
                pass
        """

        errors = [PylintError("test.py", 1, "C0111", "missing-docstring", "func", "Missing module docstring")]

        expected = """\
            # lint-amnesty, pylint: disable=missing-docstring
            def func(arg):
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_fix_module_and_func_docstring_missing(self):
        input_code = """\
            def func(arg):
                pass
        """

        errors = [
            PylintError("test.py", 1, "C0111", "missing-docstring", "func", "Missing module docstring"),
            PylintError("test.py", 1, "C0111", "missing-docstring", "func", "Missing function docstring"),
        ]

        expected = """\
            # lint-amnesty, pylint: disable=missing-docstring
            def func(arg):  # lint-amnesty, pylint: disable=missing-docstring
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_fix_many_errors(self):
        input_code = """\
            def foo(self, arg, blarg, flarg, blanth, notehu, onethu, psonatehu):
                pass"""

        errors = [
            PylintError("lint_test.py", 2, "C0304", "missing-final-newline", "", "Final newline missing"),
            PylintError("lint_test.py", 1, "C0111", "missing-docstring", "", "Missing module docstring"),
            PylintError("lint_test.py", 1, "C0102", "blacklisted-name", "foo", 'Black listed name "foo"'),
            PylintError("lint_test.py", 1, "C0111", "missing-docstring", "foo", "Missing function docstring"),
            PylintError("lint_test.py", 1, "W0613", "unused-argument", "foo", "Unused argument 'notehu'"),
            PylintError("lint_test.py", 1, "W0613", "unused-argument", "foo", "Unused argument 'self'"),
            PylintError("lint_test.py", 1, "W0613", "unused-argument", "foo", "Unused argument 'psonatehu'"),
            PylintError("lint_test.py", 1, "W0613", "unused-argument", "foo", "Unused argument 'blanth'"),
            PylintError("lint_test.py", 1, "W0613", "unused-argument", "foo", "Unused argument 'onethu'"),
            PylintError("lint_test.py", 1, "W0613", "unused-argument", "foo", "Unused argument 'arg'"),
            PylintError("lint_test.py", 1, "W0613", "unused-argument", "foo", "Unused argument 'blarg'"),
            PylintError("lint_test.py", 1, "W0613", "unused-argument", "foo", "Unused argument 'flarg'"),
        ]

        expected = """\
            # lint-amnesty, pylint: disable=missing-docstring
            def foo(self, arg, blarg, flarg, blanth, notehu, onethu, psonatehu):  # lint-amnesty, pylint: disable=blacklisted-name, missing-docstring, unused-argument
                pass  # lint-amnesty, pylint: disable=missing-final-newline"""

        self.assert_amnesty(input_code, errors, expected)



"""Plugin management for edx-lint.

This module imports all our plugins, and creates the register function that
will register them with pylint.
"""

from edx_lint.pylint import (
    annotations_check,
    getattr_check,
    i18n_check,
    module_trace,
    range_check,
    super_check,
    layered_test_check,
    right_assert_check,
    yaml_load_check,
)

MODS = [
    annotations_check,
    getattr_check,
    i18n_check,
    module_trace,
    range_check,
    super_check,
    layered_test_check,
    right_assert_check,
    yaml_load_check,
]


def register(linter):
    """Registering additional checkers.
    However, we will also use it to amend existing checker config.
    """
    # add all of the checkers
    for mod in MODS:
        mod.register_checkers(linter)



"""The edx_lint list command."""

from edx_lint.metadata import KNOWN_FILES


def list_main(argv_unused):  # pylint: disable=unused-argument
    """
    list
        List the files that edx_lint can provide.
    """
    print("edx_lint knows about these files:")
    for filename, metadata in KNOWN_FILES.items():
        if not metadata.internal_only:
            print(f"  {filename}")

    return 0



"""
Information about the files we can write.
"""

import dataclasses
import os.path


@dataclasses.dataclass
class Metadata:
    format: str
    comment: str
    internal_only: bool = False


KNOWN_FILES = {
    "pylintrc": Metadata(
        format="ini",
        comment="# {}",
    ),
    ".editorconfig": Metadata(
        format="ini",
        comment="# {}",
    ),
    "commitlint.config.js": Metadata(
        format="js",
        comment="// {}",
    ),

    "just_for_testing.txt": Metadata(
        format="txt",
        comment="-- {}",
        internal_only=True,
    ),
    # BTW: common_constraints.txt is in the resources, but isn't a file
    # that edx_lint writes.
}


def existing_known_files():
    """Yield the names of files that edx_lint writes, and that exist."""
    for filename, metadata in KNOWN_FILES.items():
        if metadata.internal_only:
            continue
        if os.path.exists(filename):
            yield filename



"""Things common to all pylint checkers."""

from astroid.nodes import ALL_NODE_CLASSES

BASE_ID = 76


CLASS_NAMES = {c.__name__.lower() for c in ALL_NODE_CLASSES}


def check_visitors(cls):
    """Check that a checker's visitors are correctly named.

    A checker has methods named visit_NODETYPE, but it's easy to mis-name
    a visit method, and it will never be called.  This decorator checks
    the class to see that all of its visitors are named after an existing
    node class.
    """
    for name in dir(cls):
        if name.startswith("visit_"):
            if name[6:] not in CLASS_NAMES:
                raise Exception(f"Method {name} doesn't correspond to a node class")
    return cls


def usable_class_name(node):
    """Make a reasonable class name for a class node."""
    name = node.qname()
    for prefix in ["__builtin__.", "builtins.", "."]:
        if name.startswith(prefix):
            name = name[len(prefix) :]
    return name



"""Test layered_test_check.py"""

from .pylint_test import run_pylint


def test_layered_test_checker():
    source = """\
        import unittest

        # This is bad

        class TestCase(unittest.TestCase):
            def test_one(self):
                pass

        class DerivedTestCase(TestCase):        #=A
            def test_two(self):
                pass

        # No big deal, the base class isn't a test.

        class TestHelpers(unittest.TestCase):
            __test__ = False
            def test_one(self):
                pass

        class TestsWithHelpers(TestHelpers):
            __test__ = True
            def test_two(self):
                pass

        # Bad: this base class is a test.

        class TestHelpers2(unittest.TestCase):
            __test__ = True
            def test_one(self):
                pass

        class TestsWithHelpers2(TestHelpers2):  #=B
            def test_two(self):
                pass

        # Mixins are fine.

        class TestMixins(object):
            def test_one(self):
                pass

        class TestsWithMixins(TestMixins, unittest.TestCase):
            def test_two(self):
                pass

        # A base class which is a TestCase, but has no test methods.

        class EmptyTestCase(unittest.TestCase):
            def setUp(self):
                super().setUp()

        class ActualTestCase(EmptyTestCase):
            def test_something(self):
                pass

        # Bizzaro __test__ examples to complete branch coverage.

        class WhatIsThis(unittest.TestCase):
            def __test__(self):
                return self.fail("I don't know what I'm doing.")

        class TooTrickyForTheirOwnGood(unittest.TestCase):
            __test__ = 1 - 1

        # The warning can be disabled.

        class TestCase(unittest.TestCase):
            def test_one(self):
                pass

        class DerivedTestCase(TestCase):        # pylint: disable=test-inherits-tests
            def test_two(self):
                pass
    """
    msg_ids = "test-inherits-tests"
    messages = run_pylint(source, msg_ids)
    expected = {
        "A:test-inherits-tests:test class DerivedTestCase inherits tests from TestCase",
        "B:test-inherits-tests:test class TestsWithHelpers2 inherits tests from TestHelpers2",
    }
    assert expected == messages



"""
edx-lint
========

A collection of code quality tools:

- A few pylint plugins to check for quality issues pylint misses.

- A command-line tool to generate config files like pylintrc from a master
  file (part of edx_lint), and a repo-specific tweaks file.

"""

import os
import re
import setuptools


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files, including any constraints from other files that
    are pulled in.
    Returns a list of requirement strings.
    """
    requirements = {}
    constraint_files = set()

    # groups "my-package-name<=x.y.z,..." into ("my-package-name", "<=x.y.z,...")
    requirement_line_regex = re.compile(r"([a-zA-Z0-9-_.]+)([<>=][^#\s]+)?")

    def add_version_constraint_or_raise(current_line, current_requirements, add_if_not_present):
        regex_match = requirement_line_regex.match(current_line)
        if regex_match:
            package = regex_match.group(1)
            version_constraints = regex_match.group(2)
            existing_version_constraints = current_requirements.get(package, None)
            # it's fine to add constraints to an unconstrained package, but raise an error if there are already
            # constraints in place
            if existing_version_constraints and existing_version_constraints != version_constraints:
                raise BaseException(f'Multiple constraint definitions found for {package}:'
                                    f' "{existing_version_constraints}" and "{version_constraints}".')
            if add_if_not_present or package in current_requirements:
                current_requirements[package] = version_constraints

    # process .in files and store the path to any constraint files that are pulled in
    for path in requirements_paths:
        with open(path) as reqs:
            for line in reqs:
                if is_requirement(line):
                    add_version_constraint_or_raise(line, requirements, True)
                if line and line.startswith('-c') and not line.startswith('-c http'):
                    constraint_files.add(os.path.dirname(path) + '/' + line.split('#')[0].replace('-c', '').strip())

    # process constraint files and add any new constraints found to existing requirements
    for constraint_file in constraint_files:
        with open(constraint_file) as reader:
            for line in reader:
                if is_requirement(line):
                    add_version_constraint_or_raise(line, requirements, False)

    # process back into list of pkg><=constraints strings
    return [f'{pkg}{version or ""}' for (pkg, version) in sorted(requirements.items())]


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement.
    Returns:
        bool: True if the line is not blank, a comment, a URL, or an included file
    """
    return line and line.strip() and not line.startswith(("-r", "#", "-e", "git+", "-c"))


def get_version(*file_paths):
    """
    Extract the version string from the file at the given relative path fragments.
    """
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    with open(filename, encoding='utf-8') as opened_file:
        version_file = opened_file.read()
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


VERSION = get_version("edx_lint", "__init__.py")


setuptools.setup(
    version=VERSION,
    include_package_data=True,
    install_requires=load_requirements("requirements/base.in"),
)



"""edx_lint unittest_assert module (optional plugin for unittest assertion checks).

Add this to your pylintrc::

    load-plugins=edx_lint.pylint.unittest_assert

"""

from .unittest_assert_check import register_checkers

register = register_checkers



"""Tests of edx_lint/pylint/common.py"""

import pytest

from edx_lint.pylint.common import check_visitors

# pylint: disable=unused-variable


def test_check_good_visitors():
    @check_visitors
    # pylint: disable=missing-class-docstring
    class ItsRight:
        def visit_call(self):
            pass  # pragma: no cover

        def this_isnt_checked(self):
            pass  # pragma: no cover

        def visit_classdef(self):
            pass  # pragma: no cover


def test_check_bad_visitors():
    msg = "Method visit_xyzzy doesn't correspond to a node class"
    with pytest.raises(Exception, match=msg):

        @check_visitors
        class ItsNotRight:
            def visit_xyzzy(self):
                pass  # pragma: no cover
